#pragma once
#include "../../stdafx.h"

#include "../glLibMain.h"

class glLibFont {
	friend void glLibInternal_init(void);
	friend void glLibInternal_deinit(void);

	typedef map<string,TTF_Font*> font_map_type;

	public:
		string name;
		unsigned char size;
	private:
		string* glLibInternal_current_text;

		static font_map_type* glLibInternal_fonts;

		#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
		static bool glLibInternal_fonts_okay;
		static string* glLibInternal_font_error;
		#endif

		static void glLibInternal_init_fonts(void) {
			#ifndef GLLIB_ERROR_MODE_REPORT_ERRORS
			TTF_Init()
			#else
			glLibInternal_fonts_okay = (TTF_Init()==0);

			glLibInternal_font_error = NULL;
			if (!glLibInternal_fonts_okay) {
				glLibInternal_font_error = new string(SDL_GetError());
			}
			#endif

			glLibInternal_fonts = new font_map_type();
		}
		static void glLibInternal_deinit_fonts(void) {
			font_map_type::const_iterator iter;
			for (iter=glLibInternal_fonts->begin();iter!=glLibInternal_fonts->end();++iter) {
				TTF_CloseFont(iter->second);
			}
			delete glLibInternal_fonts;
			if (glLibInternal_font_error) delete glLibInternal_font_error;

			TTF_Quit();
		}

		glLibTexture2D* glLibInternal_rendered_texture;

	public:
		glLibFont(char* font_name, unsigned char size) {
			glLibInternal_init(font_name,size);
			glLibInternal_load_font();
		}
		glLibFont(string font_name, unsigned char size) {
			glLibInternal_init(font_name,size);
			glLibInternal_load_font();
		}
		~glLibFont() {
			glLibInternal_deinit();
		}

		const glLibTexture2D* render(char* text, float color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text,color_struct);
		}
		const glLibTexture2D* render(char* text, double color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text,color_struct);
		}
		template <typename type_int> inline const glLibTexture2D* render(char* text, type_int color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text,color_struct);
		}
		const glLibTexture2D* render(string text, float color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text.c_str(),color_struct);
		}
		const glLibTexture2D* render(string text, double color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text.c_str(),color_struct);
		}
		template <typename type_int> inline const glLibTexture2D* render(string text, type_int color[3]) {
			if ((*glLibInternal_current_text==text)&&(glLibInternal_rendered_texture!=NULL)) return glLibInternal_rendered_texture;
			SDL_Color color_struct; glLibInternal_get_color(color,&color_struct);
			return glLibInternal_render(text.c_str(),color_struct);
		}
	private:
		void glLibInternal_get_color(float color[3], SDL_Color* color_struct) {
			color_struct->r = glLibMathMisc::rnduchar(255.0*color[0]);
			color_struct->g = glLibMathMisc::rnduchar(255.0*color[1]);
			color_struct->b = glLibMathMisc::rnduchar(255.0*color[2]);
		}
		void glLibInternal_get_color(double color[3], SDL_Color* color_struct) {
			color_struct->r = glLibMathMisc::rnduchar(255.0*color[0]);
			color_struct->g = glLibMathMisc::rnduchar(255.0*color[1]);
			color_struct->b = glLibMathMisc::rnduchar(255.0*color[2]);
		}
		template <typename type_int> void glLibInternal_get_color(type_int color[3], SDL_Color* color_struct) {
			color_struct->r = color[0];
			color_struct->g = color[1];
			color_struct->b = color[2];
		}

		const glLibTexture2D* glLibInternal_render(const char* text, SDL_Color color) {
			SDL_Surface* surf;
			surf = TTF_RenderText_Blended((*glLibInternal_fonts)[name+to_string(size)],text,color);
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (surf==NULL) throw glLibError("Could not render text ("+string(SDL_GetError())+")!");
			#endif

			if (glLibInternal_rendered_texture) delete glLibInternal_rendered_texture;
			glLibInternal_rendered_texture = new glLibTexture2D(surf,GLLIB_ALL,GLLIB_RGBA,GLLIB_FILTER_BILINEAR_MIPMAP_BILINEAR,GLLIB_FILTER_BILINEAR);

			SDL_FreeSurface(surf);

			return glLibInternal_rendered_texture;
		}

		void glLibInternal_init(string new_name, int new_size) {
			name = new_name;
			size = new_size;

			glLibInternal_rendered_texture = NULL;

			glLibInternal_current_text = new string("");
		}
		void glLibInternal_deinit(void) {
			if (glLibInternal_rendered_texture) delete glLibInternal_rendered_texture;
			delete glLibInternal_current_text;
		}

		void glLibInternal_load_font(void) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (!glLibInternal_fonts_okay) {
				throw glLibError("Fonts not available ("+(*glLibInternal_font_error)+")!");
			}
			#endif

			string key_name = name + to_string(size);

			font_map_type::const_iterator iter = glLibInternal_fonts->find(key_name);
			if (iter==glLibInternal_fonts->end()) {
				TTF_Font* font = TTF_OpenFont(name.c_str(),size); //TODO: test/handle failure
				(*glLibInternal_fonts)[key_name] = font;
			}
		}
};